/*
 * Styles for navigation
 */

#slider {
    display: none;
    position: fixed;
    top: 106px;
    bottom: 0;
    left: 0;
    z-index: 5;
    width: 50%;
    min-width: 400px;
    background-color: #f8f8f8;
    border: 1px solid rgba(10, 45, 105, 0.8);
}

#slider-tabs, #slider .tab-content {
    flex: auto;
    -ms-flex: auto;
    -moz-flex: auto;
    -webkit-flex: auto;
    height: 100%;
}

#slider-tabs .active {
    background-color: #ffffff;
}

#slider-tabs .tab-pane {
    padding: 10px;
    overflow: auto;
    position: absolute;
    top: 43px;
    bottom: 0;
    width: 100%;
}

#slider-filter-box {
    float: right;
    padding: 7px;
}

.sidebar-button-main {
    position: fixed;
    left: 0;
    top: 208px;
}

.sidebar-button-slider {
    position: absolute;
    right: -40px;
    top: 100px;
}

.sidebar-button {
    cursor: pointer;
    color: white;
    z-index: 5;
    box-shadow: 2px 2px 1px #cecece;
    height: 40px;
    opacity: 0.5;
    width: 40px;
    background-color: #0a2d69;
    display: inline-block;
    font-size: 1.7em;
    line-height: 1.6;
    text-align: center;
    border: medium none;
    border-top-right-radius: 50%;
    border-bottom-right-radius: 50%;
}

.slider-option {
    cursor: pointer;
    position: relative;
}

.slider-option:hover {
    background-color: #e4e4e4;
}

.slider-option-name {
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
}

.slider-option-anno {
    color: gray;
    position: absolute;
    right: 15px;
    top: 10px;
}

.slider-option-desc {
    color: #333366;
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
    display: block;
}

.slider-option-tags {
    display: block;
    color: rgba(10, 45, 105, 1);
    padding: 5px;
}

/*
 * Styles for specific GenePattern Public Server
 */

.nbtools-widget.gp-widget.gp-server-public {
    border-color: rgba(10, 45, 105, 0.80);
}

.nbtools-widget.gp-widget.gp-server-public .panel-heading {
    background-color: rgba(10, 45, 105, 0.80);
}

.nbtools-widget.gp-widget.gp-server-public .btn-primary {
    border-color: rgba(10, 45, 105, 0.80);
    background-color: rgba(10, 45, 105, 0.80);
}

.nbtools-widget.gp-widget.gp-server-public .btn-primary:hover {
    border-color: rgba(10, 45, 105, 1);
    background-color: rgba(10, 45, 105, 1);
}

/*
 * Styles for specific GenePattern Indiana Server
 */

.nbtools-widget.gp-widget.gp-server-indiana {
    border-color: rgba(115, 25, 10, 0.80);
}

.nbtools-widget.gp-widget.gp-server-indiana .panel-heading {
    background-color: rgba(115, 25, 10, 0.80);
}

.nbtools-widget.gp-widget.gp-server-indiana .btn-primary {
    border-color: rgba(115, 25, 10, 0.80);
    background-color: rgba(115, 25, 10, 0.80);
}

.nbtools-widget.gp-widget.gp-server-indiana .btn-primary:hover {
    border-color: rgba(115, 25, 10, 1);
    background-color: rgba(115, 25, 10, 1);
}

/*
 * Styles for specific GenePattern Broad Server
 */

.nbtools-widget.gp-widget.gp-server-broad {
    border-color: rgba(15, 75, 105, 0.80);
}

.nbtools-widget.gp-widget.gp-server-broad .panel-heading {
    background-color: rgba(15, 75, 105, 0.80);
}

.nbtools-widget.gp-widget.gp-server-broad .btn-primary {
    border-color: rgba(15, 75, 105, 0.80);
    background-color: rgba(15, 75, 105, 0.80);
}

.nbtools-widget.gp-widget.gp-server-broad .btn-primary:hover {
    border-color: rgba(15, 75, 105, 1);
    background-color: rgba(15, 75, 105, 1);
}

/*
 * Styles for specific GenePattern Custom Server
 */

.nbtools-widget.gp-widget.gp-server-custom {
    border-color: rgba(15, 105, 75, 0.80);
}

.nbtools-widget.gp-widget.gp-server-custom .panel-heading {
    background-color: rgba(15, 105, 75, 0.80);
}

.nbtools-widget.gp-widget.gp-server-custom .btn-primary {
    border-color: rgba(15, 105, 75, 0.80);
    background-color: rgba(15, 105, 75, 0.80);
}

.nbtools-widget.gp-widget.gp-server-custom .btn-primary:hover {
    border-color: rgba(15, 105, 75, 1);
    background-color: rgba(15, 105, 75, 1);
}

/*
 * Local execution in GenePattern
 */

.nbtools-widget.gp-widget.gp-server-local {
    border-color: rgba(43, 43, 43, 0.8);
}

.nbtools-widget.gp-widget.gp-server-local .panel-heading {
    background-color: rgba(43, 43, 43, 0.8);
}

.nbtools-widget.gp-widget.gp-server-local .btn-primary {
    border-color: rgba(43, 43, 43, 0.8);
    background-color: rgba(43, 43, 43, 0.8);
}

.nbtools-widget.gp-widget.gp-server-local .btn-primary:hover {
    border-color: rgba(43, 43, 43, 1);
    background-color: rgba(43, 43, 43, 1);
}

/*
 * Styles for widget display
 */

.widget-float-right {
    float: right;
    position: relative;
    top: -3px;
    left: 7px;
}

.nbtools-widget.gp-widget {
    width: 100%;
    max-width: 100%;
    border-color: rgba(10, 45, 105, 0.80);
    overflow: visible !important;
}

.nbtools-widget.gp-widget.panel {
    margin-bottom: 0;
}

.nbtools-widget.gp-widget .panel-heading {
    background-color: rgba(10, 45, 105, 0.80);
    color: white;
}

.nbtools-widget.gp-widget label {
    font-weight: bold;
}

.gp-widget-logo {
    height: 20px !important;
    float: left;
    padding-right: 10px;
    position: relative;
    left: -5px;
    top: -2px;
}

.widget-username-label {
    font-style: italic;
}

.widget-server-label {
    font-size: 0.8em;
    font-style: italic;
    padding-right: 10px;
}

.nbtools-widget.gp-widget .btn-primary {
    border-color: rgba(10, 45, 105, 0.80);
    background-color: rgba(10, 45, 105, 0.80);
}

.nbtools-widget.gp-widget .btn-primary:hover {
    border-color: rgba(10, 45, 105, 1);
    background-color: rgba(10, 45, 105, 1);
}

.widget-view {
    position: relative;
}

.widget-subarea {
    max-width: 100%;
}

div.output_subarea {
    overflow-x: visible;
}

div.cell.selected {
    z-index: 10;
}

.widget-auto-login {
    position: absolute;
    top: 0;
    bottom: 0;
    left: 0;
    right: 0;
    background: rgba(255,255,255,0.8);
}

.widget-auto-login .panel {
    width: 50%;
    position: absolute;
    top: 50%;
    left: 50%;
    transform: translate(-50%,-50%);
}

.widget-auto-login-buttons {
    text-align: center;
    margin-top: 10px;
}

.gp-widget-loading, .gp-widget-logged-in {
    position: absolute;
    text-align: center;
    top: 0;
    bottom: 0;
    right: 0;
    left: 0;
    background: rgba(255,255,255,0.75);
}

.gp-widget-loading img {
    width: 100px;
    position: relative;
    top: 75px;
}

.gp-widget-logged-in div {
    position: relative;
    top: 75px;
    display: inline-block;
    line-height: 40px;
    font-size: 1.2em;
}

.widget-code {
    -moz-box-align: stretch;
    -moz-box-orient: horizontal;
    align-items: stretch;
    display: flex;
    flex-direction: row;
    page-break-inside: avoid;
}

.widget-code .prompt {
    display: inline-block;
}

.widget-code pre {
    border: none;
    border-radius: 0;
    padding: 0;
    margin: 0;
}

/*
 * Job Widget Styles
 */

.gp-widget-job {
    position: relative;
    top: -5px;
    border-radius: 0 0 2px 2px;
}

.gp-widget-job-header {
    /*display: none;*/
    opacity: 0.6;
}

.gp-widget-job-header > img {
    display: none;
}

.gp-widget-job .gp-widget-job-task {
    display: none;
}

.gp-widget-job-float-right {
    float: right;
    position: relative;
    z-index: 1;
    top: 6px;
    right: 7px;
}

.output_area:first-child .gp-widget-job {
    /* Handle the style when it is a standalone widget */
    position: static;
    top: auto;
    border-radius: inherit;
}

.output_area:first-child .gp-widget-job > .gp-widget-job-header {
    /* Handle the style when it is a standalone widget */
    display: block;
    opacity: 1;
}

.output_area:first-child .gp-widget-job .gp-widget-job-task {
    /* Handle the style when it is a standalone widget */
    display: inline-block;
}

.output_area:first-child .gp-widget-job-header > img {
    /* Handle the style when it is a standalone widget */
    display: block;
}

.gp-widget-job-status {
    padding-bottom: 3px;
    text-align: right;
    padding-right: 20px;
}

.gp-widget-job-submitted {
    font-style: italic;
    font-size:0.9em;
}

.gp-widget-job-sharing {
    display: inline-block;
    margin-left: 10px;
    cursor: pointer;
}

.gp-widget-job-buttons {
    text-align: right;
    display: inline-block;
}

.gp-widget-job-share-table {
    width: 100%;
}

.gp-widget-job-share-table input[type=radio] {
    margin-left: 20px;
    margin-right: 5px;
}

.gp-widget-job-child {
    border-color: #BBBBBB !important;
}

.gp-widget-job-child .gp-widget-job-header {
    background-color: #BBBBBB !important;
    border-color: #BBBBBB !important;
}

.gp-widget-job-outputs-list .popover label {
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
}

/*
 * Task & UI Builder Styles
 */

.gp-widget-task > .panel-body,
.gp-widget-call > .panel-body {
    padding: 0;
}

.gp-widget-task > .panel-body > .alert {
    margin-top: 10px;
    margin-right: 10px;
    margin-left: 10px;
}

.gp-widget-task-version {
    font-size: 0.8em;
    font-style: italic;
    padding-right: 10px;
}

.gp-widget-task-subheader {
    border-bottom: 1px solid rgba(10, 45, 105, 0.8);
    /*background: #d3d3d355;*/
    padding-top: 10px;
    padding-left: 10px;
    padding-right: 10px;
}

.gp-widget-task-desc {
    padding-bottom: 10px;
}

.gp-widget-task-run {
    overflow: auto;
    margin-left: 20px;
    font-size: 0.9em;
    padding-top: 10px;
    float: right;
}

.gp-widget-task-run-button {
    float: right;
    margin-bottom: 5px;
}

.gp-widget-task-param {
    margin: 15px;
    padding-right: 20px;
}

.gp-widget-task-param-name {
    display: inline-block;
    min-width: 150px;
    font-weight: bold;
    font-size: 0.9em;
    vertical-align: top;
    padding-top: 10px;
    padding-right: 10px;
}

.gp-widget-task-param-wrapper {
    vertical-align: top;
    padding-left: 10px;
}

.gp-widget-task-param-desc {
    font-size: 0.9em;
    margin-top: 5px;
}

.gp-widget-task-param-missing {
    border: 2px solid red;
}

.gp-widget-task-footer {
    border-top: 1px solid rgba(10, 45, 105, 0.8);
    padding-top: 10px;
    /*background: #d3d3d355;*/
    padding-left: 10px;
    padding-right: 10px;
}

.gp-widget-task-eula-license {
    width: 100%;
    height: 300px;
    overflow: auto;
    text-align: left;
    font-size: 0.8em;
    font-weight: normal;
    line-height: 1.2em;
}

.gp-widget-task-eula {
    padding: 20px;
}

.gp-widget-task-eula div {
    width: 100%;
    font-weight: bold;
    line-height: 3em;
}

.gp-widget-task-group-title {
    opacity: 0.75;
    margin-left: 10px;
    margin-right: 10px;
    margin-top: 10px;
}

/*
 * Typeahead Styles
 */

.gp-widget-typeahead {
    display: inline-block;
}

.gp-widget-typeahead-arrow {
    font-size: 20px;
    line-height: 32px;
    cursor: pointer;
}

.gp-widget-typeahead-input {
    display: inline-block;
}

.gp-widget-typeahead-group {
    margin-bottom: 0;
}

.gp-widget-typeahead-list {
    min-width: 100%;
    margin-top: -1px;
    max-height: 300px;
    overflow-y: auto;
    overflow-x: hidden;
}

.gp-widget-typeahead-list .dropdown-header {
    font-weight: bold;
}

.gp-widget-typeahead-list .dropdown-file {
    padding-left: 40px;
}

/*
 * Module Bundler Styles
 */

.gp-widget-module-previous-button, .gp-widget-module-next-button {
    float: right;
    margin-left: 5px;
    position: relative;
    top: -5px;
}